/*
  ==============================================================================

    UARTOutput.cpp
    Created: 15 Nov 2012 10:02:19pm
    Author:  matty

  ==============================================================================
*/
#include "UARTOutput.h"

#include <boost/asio.hpp>
#include <boost/bind.hpp>
using namespace::boost::asio;

#if _MSC_VER >= 1600
	#pragma warning( disable : 4127 )   // conditional expression is constant
#endif


UARTOutput::UARTOutput()
	: Thread("UART")
{
	setName("UART output");

	uartQueue.Create(MAX_CHARS_TO_DISPLAY * 2);

	io       = nullptr;
	port     = nullptr;	

	monofont.setHeight(16.0f);
    monofont.setTypefaceName (Font::getDefaultMonospacedFontName());
}


void UARTOutput::setEngines(PowerMeter* portEng, PowerMeter* stbdEng)
{
	portEngineRef = portEng;
	stbdEngineRef = stbdEng;
}


bool UARTOutput::Connect()
{
	Disconnect();

	if (portName.trim().length() == 0) {
		AlertWindow::showMessageBoxAsync(AlertWindow::WarningIcon, "Bad port; can't connect", "Please select a port from the combo box, or click 'Add port...'");
		return false;
	}

	try {		
		io   = new boost::asio::io_service;
		wait(250);
		port = new boost::asio::serial_port(*io, static_cast<const char*>(portName.toUTF8() ) );
	}
	catch(boost::exception const& /* e */) {
		AlertWindow::showMessageBoxAsync(AlertWindow::WarningIcon, "Port ERROR", "1) Verify correct port is selected.\n2) If device is unplugged, please plug in device.\n3) Disconnect and reconnect COM serial device.");
		return false;
	}
	
	// TODO: shoud this be before try/catch block?
	serial_port_base::baud_rate BAUD(9600);	
	// serial_port_base::character_size CSIZE(8);	
	serial_port_base::flow_control FLOW( serial_port_base::flow_control::none );	
	serial_port_base::parity PARITY( serial_port_base::parity::none );	
	serial_port_base::stop_bits STOP( serial_port_base::stop_bits::one );

	// create the I/O service that talks to the serial device
	// io_service io;
	// create the serial device, note it takes the io service and the port name
	//serial_port port( io, PORT );	

	// go through and set all the options as we need them
	// all of them are listed, but the default values work for most cases
	port->set_option( BAUD );
	//port->set_option( CSIZE );
	port->set_option( FLOW );
	port->set_option( PARITY );
	port->set_option( STOP );

	//RunManual();
	
	//port.close();
	//port.open(PORT);
	return true;
}


bool UARTOutput::RunManual()
{
	if (io == nullptr || port == nullptr) return false;
	
	startThread();
	
	// start motor control timer
	startTimer(250);
	
	return true;
}

void UARTOutput::StopManual()
{
	stopTimer();
	
	if (isThreadRunning() ) {
		signalThreadShouldExit();
		notify();
		waitForThreadToExit(2000);
		stopThread(1000);
	}
}


void UARTOutput::Disconnect()
{
	StopManual();

	try {
		if (port != nullptr) {
			port->cancel();
			port->close();
			delete port;
			port = nullptr;
		}
		if (io   != nullptr) {
			io->stop();
			delete io;
			io = nullptr;
		}
	}
	catch(boost::exception const&) {
		jassert(false);
		AlertWindow::showMessageBoxAsync(AlertWindow::WarningIcon, "Port ERROR", "Unable to stop COM serial device."); 		
		return;
	}
}


void UARTOutput::SetPortName(const String& name)
{
	portName = name;
}

UARTOutput::~UARTOutput()
{	
	Disconnect();
}


void UARTOutput::paint(Graphics& g)
{
	g.drawImageAt(uartDisplay, 0, 0);
	notify();
	//if (!isThreadRunning() ) notify();
}


void UARTOutput::render()
{
	static char copy[4096];

	//String str;
	unsigned int len;

	Graphics g(uartDisplay);

	// make copy of string for rendering	 
	{ // BEGIN CRITICAL SECTION
		const ScopedReadLock myScopedLock (uartQueueLock);
		len = uartQueue.Copy(copy, 4095);
	} // END CRITICAL SECTION

	g.fillAll(Colours::black);	
	g.setFont(monofont);
	g.setColour(Colours::white);

	GlyphArrangement ga;

	int xpos = 0;
	int ypos = 16;
		
	//const char* cstr = static_cast<const char*>(str.toUTF8() );
	char n[2];
	n[1] = 0;

	for (unsigned int i = 0; i < len; i++) {
		if (copy[i] == '\r') {
			ga.draw(g);
			ga.clear();
			g.setColour(Colours::red);
			continue;
		}
		else if (copy[i] == '\t') {
			ga.draw(g);
			ga.clear();
			g.setColour(Colours::lime);
			continue;
		}
		
		n[0] = copy[i];

		int len = monofont.getStringWidth( String(n) );
		if (xpos + len > getWidth() ) {
			xpos = 0;
			ypos += 16;
		}
		ga.addCurtailedLineOfText(monofont, String(n), static_cast<float>(xpos), static_cast<float>(ypos), static_cast<float>(getWidth()), false);
		xpos += len;
	} // end for

	ga.draw(g);

}

void UARTOutput::send_string(const String& str)
{
	if (port == nullptr || io == nullptr) {
		AlertWindow::showMessageBoxAsync(AlertWindow::WarningIcon, "Send ERROR", "Unable to write serial data.\nPlease make sure the device is plugged in.");
		return;
	}

	char ascii[128];
	int lengthIncludingNull = str.copyToUTF8(ascii, 128);
	char n[2];
	n[1] = 0;

	try {
		for (int i = 0; i < lengthIncludingNull - 1; i++) {
			//n[0] = reverseBitOrder(pcstr[0]);
			n[0] = ascii[i];
			//write(*port, buffer(n, 1) );
			port->write_some(buffer(n, 1) );
		}
	} catch (...) {
		Disconnect();
		AlertWindow::showMessageBoxAsync(AlertWindow::WarningIcon, "Send ERROR", "Unable to write serial data.\nWas the device unplugged?");
	}
}

void UARTOutput::Send(String str)
{
	// Send -> UART //////////////////////
	send_string(str);
	
	// UPDATE GUI ////////////////////////
	
	// prefix with \r so it gets color coded
	str = L"\r" + str;
	
	{ // BEGIN CRITICAL SECTION
		const ScopedReadLock myScopedLock (uartQueueLock);
		
		int overflow = uartQueue.NumElementsInQueue() - uartQueue.CountSpecialChars() + str.length() - 1 - MAX_CHARS_TO_DISPLAY;
		if (overflow > 0) uartQueue.Pop(overflow);
		
		uartQueue.Push(static_cast<const char*>(str.toUTF8() ), str.length());
	} // END CRITICAL SECTION
	
	render();
	repaint();	
}

void UARTOutput::SendText(TextEditor & t)
{
	if (t.getText().length() == 0) return;

	Send(t.getText() );
}

void UARTOutput::handleAsyncUpdate()
{
	render();
	repaint();	
}

void UARTOutput::resized()
{	
	if (uartDisplay.getWidth() > 0) {
		jassert(false);  // we only allow creation once!
		return;
	}

	uartDisplay = Image(Image::ARGB, getWidth(), getHeight(), true);
	
	Graphics g(uartDisplay);
	g.fillAll(Colours::black);
}

//void UARTOutput::log(const String& message)
//{
//	{  		
//		const MessageManagerLock mmLock;
//		if (!mmLock.lockWasGained() ) {		
//			return;
//		}
//				
//		// editor->insertTextAtCaret(message);
//	} // the lock is released here	
//}

unsigned char UARTOutput::reverseBitOrder(unsigned char c)
{		
	unsigned char n = 0;

	// 1
	n <<= 1;
	n |= c & 1;			
	c >>= 1;
	
	// 2
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 3
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 4
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 5
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 6
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 7
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	// 8
	n <<= 1;
	n |= c & 1;			
	c >>= 1;

	return n;
}


void UARTOutput::run()
{
	char n[2];
	n[1] = 0; // null terminate
	
	if (io == nullptr) return;

	deadline_timer timeout(*io);
	bool data_available;

	while(!threadShouldExit() )
	{
		// http://stackoverflow.com/questions/456042/how-do-i-perform-a-nonblocking-read-using-asio
		port->async_read_some(buffer(&readBuffer, MAX_CHARS_TO_DISPLAY),
		                      boost::bind(&UARTOutput::read_callback, this,
		                                  boost::ref(data_available),
								 	      boost::ref(timeout),
		                                  placeholders::error,
		                                  placeholders::bytes_transferred)
		);
		
		//timeout.expires_from_now(boost::posix_time::milliseconds(250));
		timeout.expires_from_now(boost::posix_time::milliseconds(250));

		
		timeout.async_wait(boost::bind(&UARTOutput::wait_callback, this,
		                               boost::ref(*port),
		                               placeholders::error)
		);

		// will block until async callbacks are finished
		io->run();
		io->reset();

		if (data_available) {
			triggerAsyncUpdate();			
			wait(-1);			
		}
		
	} // while()

} // run()


void UARTOutput::read_callback(bool& data_available, deadline_timer& timeout, const boost::system::error_code& error, std::size_t bytes_transferred)
{
	if (error || !bytes_transferred)
	{
		// No data was read!
		data_available = false;
		return;
	}

	timeout.cancel();  // will cause wait_callback to fire with an error
	data_available = true;

	
	////////////////////////////////////
	
	char rc;
	unsigned int i;
	for(i = 0; i < bytes_transferred; i++) {
		//rc = UARTOutput::reverseBitOrder(readBuffer[i]);
		rc = readBuffer[i];
		// check range to make juce happy
		if (rc > 127 || rc <= 0) rc = 127;
		//else if (rc == 0) rc = 127; // null is end of string marker
		else if (rc == '\r') rc = 127; // special receive color marker
		readBuffer[i] = rc;
	}
	readBuffer[i] = 0; // null terminate

	{ // BEGIN CRITICAL SECTION
		const ScopedReadLock myScopedLock (uartQueueLock);

		int overflow = uartQueue.NumElementsInQueue() - uartQueue.CountSpecialChars() + bytes_transferred - MAX_CHARS_TO_DISPLAY;
		if (overflow > 0) uartQueue.Pop(overflow);

		uartQueue.Push('\t');
		uartQueue.Push(readBuffer, bytes_transferred);
	} // END CRITICAL SECTION

}

void UARTOutput::wait_callback(serial_port& ser_port, const boost::system::error_code& error)
{
  if (error)
  {
    // Data was read and this timeout was canceled
    return;
  }

  ser_port.cancel();  // will cause read_callback to fire with an error
}

String UARTOutput::padZeros(String str)
{
	return str.paddedLeft('0', 3);
}

void UARTOutput::timerCallback()
{	
	float port = (128.0 - portEngineRef->getValue() ) * portEngineRef->getRight() * 2;
	if (port > 255.0) port = 255.0;
	else if (port < -255.0) port = -255.0;
	
	float stbd = (128.0 - portEngineRef->getValue() ) * portEngineRef->getLeft() * 2;
	if (stbd > 255.0) stbd = 255.0;
	else if (stbd < -255.0) stbd = -255.0;
	
	if (port >= 0)     Send("$MOT1<" + padZeros(String(static_cast<int>(port))) + ";");
	else if (port < 0) Send("$MOT1>" + padZeros(String(-static_cast<int>(port))) + ";");
	
	if (stbd >= 0)     Send("$MOT0<" + padZeros(String(static_cast<int>(stbd))) + ";");
	else if (stbd < 0) Send("$MOT0>" + padZeros(String(-static_cast<int>(stbd))) + ";");
	
}

